(*---------------------------------------------------------------------------
   Copyright (c) 2014 The bos programmers. All rights reserved.
   Distributed under the ISC license, see terms at the end of the file.
  ---------------------------------------------------------------------------*)

open Rresult

(* Basic types *)

module Pat = Bos_pat
module Cmd = Bos_cmd

(* OS interaction *)

module OS = struct
  type ('a, 'b) result = ('a, [> R.msg] as 'b) R.t
  module Env = Bos_os_env
  module Arg = Bos_os_arg
  module Path = Bos_os_path
  module File = Bos_os_file
  module Dir = Bos_os_dir
  module Cmd = Bos_os_cmd
  module U = Bos_os_u
end

(*---------------------------------------------------------------------------
   Copyright (c) 2014 The bos programmers

   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
   copyright notice and this permission notice appear in all copies.

   THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
   WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
   ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
   ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  ---------------------------------------------------------------------------*)
